/*
 * Copyright (C) 2002  Pekka Enberg <Pekka.Enberg@cs.helsinki.fi>
 *
 * Distributed under the terms of the GNU General Public License
 * version 2 or later.
 */
#include "xmlindent.h"
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DEFAULT_INDENT_CHAR ' '
#define DEFAULT_NUM_INDENT_CHARS 4

char indent_char = DEFAULT_INDENT_CHAR;
int num_indent_chars = DEFAULT_NUM_INDENT_CHARS;
bool leave_comments_alone = false;

/* Print error message.  */
void error(const char * format, ...)
{
    va_list args;

    fprintf(stderr, "Error: ");

    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    fputc('\n', stderr);
}

/* Print indent characters.  */
void print_indent(FILE * output, int indent_level)
{
    int i;

    for (i = 0; i < (num_indent_chars * indent_level); i++) {
	fputc(indent_char, output);
    }
}

/* Buffer used to hold XML stream.  */
struct buffer {
    int * data;
    size_t length;
    size_t capacity;
};

/* Initialize buffer.  */
void buffer_init(struct buffer * buffer, size_t initial_capacity)
{
    buffer->data     = malloc(initial_capacity * sizeof(int));
    buffer->length   = 0;
    buffer->capacity = initial_capacity;
}

/* Release buffer.  */
void buffer_release(struct buffer * buffer)
{
    if (buffer->capacity > 0) {
	free(buffer->data);
	buffer->data     = NULL;
	buffer->capacity = 0;
    }
    buffer->length = 0;
}

/* Check if buffer is full.  */
bool buffer_is_full(struct buffer * buffer)
{
    return (buffer->capacity == buffer->length);
}

/* Grow buffer by one.  */
void buffer_grow(struct buffer * buffer)
{
    buffer->data = realloc(buffer->data, buffer->capacity + 1);
    if (buffer->data == NULL) {
	error("Could not allocate memory for buffer.");
	exit(1);
    }
    buffer->capacity++;
}

/* Push item at back of buffer.  */
void buffer_push(struct buffer * buffer, int item)
{
    if (buffer_is_full(buffer)) {
	buffer_grow(buffer);
    }
    buffer->data[buffer->length] = item;
    buffer->length++;
}

/* Pop item from back of buffer.  */
int buffer_pop(struct buffer * buffer)
{
    int result = -1;

    if (buffer->length > 0) {
	buffer->length--;
	result = buffer->data[buffer->length];
    } else {
	error("Buffer underflow.");
    }
    return result;
}

/* Flush buffer to output stream.  */
void buffer_flush(struct buffer * buffer, FILE * output)
{
    int i;

    for (i = 0; i < buffer->length; i++) {
	fputc(buffer->data[i], output);
    }
    buffer->length = 0;
}

/* Check if for XML start-tag.  */
static inline bool is_xml_start_tag(int current, int lookahead)
{
    return (('<' == current) &&
	    (isalpha(lookahead) || ('_' == lookahead) || (':' == lookahead)));
}

/* Check for XML end-tag.  */
static inline bool is_xml_end_tag(int current, int lookahead)
{
    return ((('<' == current) && ('/' == lookahead)) ||
	    (('/' == current) && ('>' == lookahead)));
}

/* Check if input matches token.
   Note: Only three first characters are actually checked.  */
static inline bool match(int current, int lookahead,
			 FILE * input, const char * token)
{
    bool result = false;

    if ((token[0] == current) && (token[1] == lookahead)) {
	int tmp = fgetc(input);
	if (token[2] == tmp) {
	    result = true;
	}
	if (ungetc(tmp, input) != tmp) {
	    error("Could not restore XML stream.");
	}
    }
    return result;
}

/* Check if input matches token.  */
static inline bool match_backtrack(int current, int lookahead,
				   struct buffer * buffer,
				   const char * token)
{
    bool result = false;

    if ((token[strlen(token) - 2] == current) &&
	(token[strlen(token) - 1] == lookahead)) {
	int tmp = buffer_pop(buffer);

	if (token[strlen(token) - 3] == tmp) {
	    result = true;
	}
	buffer_push(buffer, tmp);
    }
    return result;
}

/* Check for XML comment start.  */
static inline bool is_xml_comment_start(int current, int lookahead,
					FILE * input)
{
    return match(current, lookahead, input, "<!--");
}

/* Check for XML comment end.  */
static inline bool is_xml_comment_end(int current, int lookahead,
				      struct buffer * buffer)
{
    return match_backtrack(current, lookahead, buffer, "-->");
}

/* Check for CDATA section start.  */
static inline bool is_cdata_section_start(int current, int lookahead,
					  FILE * input)
{
    return match(current, lookahead, input, "<![CDATA[");
}

/* Check for CDATA section end.  */
static inline bool is_cdata_section_end(int current, int lookahead,
					struct buffer * buffer)
{
    return match_backtrack(current, lookahead, buffer, "]]>");
}

/* XML end of line characters.  */
#define CARRIAGE_RETURN 0x0D
#define LINE_FEED       0x0A
#define NEL             0x85

static inline bool is_end_of_line(int current)
{
    if ((CARRIAGE_RETURN == current) ||
	(LINE_FEED == current) ||
	(NEL == current))
	return true;

    return false;
}

/* Eat whitespace from stream.  */
void eat_whitespace(FILE * input)
{
    for (;;) {
	int current = fgetc(input);
	if ((isspace(current) == 0) || is_end_of_line(current)) {
	    if (ungetc(current, input) != current) {
		error("Could not restore XML stream.");
	    }
	    break;
	}
    }
}

#define BUFFER_INITIAL_CAPACITY 1024

/* Indent XML stream.  */
void indent(FILE * input, FILE * output)
{
    struct buffer buffer;

    int indent_level;
    int lookahead;
    int start_tag_count;
    bool inside_comment;
    bool inside_cdata_section;

    buffer_init(&buffer, BUFFER_INITIAL_CAPACITY);

    indent_level         = 0;
    lookahead            = fgetc(input);
    start_tag_count      = 0;
    inside_comment       = false;
    inside_cdata_section = false;

    for (;;) {
	int current = lookahead;

	if (EOF == current) {
	    break;
	}
	lookahead = fgetc(input);

	if (!inside_comment) {
	    if (is_xml_start_tag(current, lookahead)) {
		/* XML start-tag changes indent level only after we have
		   finished the current line.  */
		start_tag_count++;
	    } else if (is_xml_end_tag(current, lookahead)) {
		if (start_tag_count > 0) {
		    /* Start-tag and end-tag on the same line balance each
                       other.  */
		    start_tag_count--;
		} else {
		    indent_level--;
		    if (indent_level < 0) {
			error("XML start-tags and end-tags are not balanced.");
			indent_level = 0;
		    }
		}
	    } else if (is_xml_comment_start(current, lookahead, input)) {
		inside_comment = true;
	    } else if (is_cdata_section_start(current, lookahead, input)) {
		inside_cdata_section = true;
	    } else if (is_cdata_section_end(current, lookahead, &buffer)) {
		inside_cdata_section = false;
	    }
	} else if (is_xml_comment_end(current, lookahead, &buffer)) {
	    inside_comment = false;
	}

	buffer_push(&buffer, current);

	if (is_end_of_line(current)) {
	    if ((inside_comment && leave_comments_alone)
		|| inside_cdata_section) {
		continue;
	    }
	    print_indent(output, indent_level);
	    buffer_flush(&buffer, output);

	    if (start_tag_count > 0) {
		indent_level += start_tag_count;
		start_tag_count = 0;
	    }

	    if (ungetc(lookahead, input) != lookahead) {
		error("Could not restore XML stream.");
	    }
	    eat_whitespace(input);
	    lookahead = fgetc(input);
	}
    }

    buffer_release(&buffer);
}
